/*----- Multiple Object Tracking with Linear Programming ----*/
/*----- Hao Jiang  (May at UBC) ----*/ 
/* For details, please refer to CVPR07 paper (H. Jiang, S. Fels and J. Little). */
/* Please do not distribute or use the code for non-research purposes        */ 
/* without the permission of the authors of the code.                         */ 

#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <resolv.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <cv.h>
#include <highgui.h>
#include <pthread.h>
#include "camera.h"

IplImage *gray_background[10]; 
int load_background_images();
void release_background_images();
int background_subtraction_histogram(IplImage * input_image, 
      int width, int height, int cam_num,
      int buf_num, IplImage * gray_background);

int     thread_ids[3] = {0,1,2};
int     write_buf_id = 0;
int     read_buf_id = 1;

int     write_obuf_id = 0;
int     read_obuf_id = 1;

const int frame_buffer_len = 20; // or set to 20
IplImage * video_buf[10][2][frame_buffer_len];
IplImage * display_buf[10][2][frame_buffer_len];

void *write_buf(void *idp);
void *read_buf(void *idp);
void *read_obuf(void *idp);
const int width = 320;
const int height = 240;

const int cam_num = 1;

void track(int cam_id);
void draw_bounding_boxes(int cam_id);

Camera *netcam[4];

int main(int Count, char *Strings[])
{  

    load_background_images();    

    netcam[0] = new NetworkCamera(width, height, "192.168.1.3");
    cvNamedWindow("video1", 0);
    cvResizeWindow("video1", 640, 480);

    pthread_t threads[3];
    pthread_attr_t attr;

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

    int n, m, i;
    for(i = 0; i < cam_num; i++)
    for(m = 0; m < 2; m++)
    for(n = 0; n < frame_buffer_len; n++)
    {
           video_buf[i][m][n] = cvCreateImage( cvSize(width, height), 8, 3);
           display_buf[i][m][n] = cvCreateImage( cvSize(width, height), 8, 3);
    }


   while(1)
   {
    pthread_create(&threads[0], &attr, read_buf, (void *)&thread_ids[0]);
    pthread_create(&threads[1], &attr, write_buf, (void *)&thread_ids[1]);
    pthread_create(&threads[2], &attr, read_obuf, (void *)&thread_ids[2]);

    pthread_join( threads[0], NULL);
    pthread_join( threads[1], NULL); 
    pthread_join( threads[2], NULL); 
   }

   for(i = 0; i < cam_num; i++)
   for(m = 0; m < 2; m++)
   for(n = 0; n < frame_buffer_len; n++)
   {
           cvReleaseImage(&video_buf[i][m][n]);
           cvReleaseImage(&display_buf[i][m][n]);
   }

   pthread_exit(NULL);
   release_background_images();
   delete [] netcam;
   return 0;
}

void *write_buf(void *idp)
{
    int n, m;
    for( n = 0; n < frame_buffer_len; n ++)
    {
      for( m = 0; m < cam_num; m++ )
        netcam[m]->capture_image(video_buf[m][write_buf_id][n]);
    }

    write_buf_id = 1 - write_buf_id;
    return 0;
}


void *read_buf(void *idp)
{

    int m, n; 
    for( int buf_num = 0; buf_num < frame_buffer_len; buf_num ++)
    {
        for( n = 0; n < cam_num; n++)
          cvCopy(video_buf[n][read_buf_id][buf_num], display_buf[n][write_obuf_id][buf_num]);
    }

    for(n = 0; n < cam_num; n++)
    { 
      for( int buf_num = 0; buf_num < frame_buffer_len; buf_num ++)
      { 
        background_subtraction_histogram(video_buf[n][read_buf_id][buf_num], width, height, n, buf_num, gray_background[n]);
      }

      track(n);
    } 

    read_buf_id = 1 - read_buf_id;
    write_obuf_id = 1 - write_obuf_id;
    return 0;
}

void draw_bounding_boxes(int cam_id)
{
   CvPoint pt1, pt2;
   for(int buf_num = 0; buf_num < frame_buffer_len; buf_num ++)
   {
       char file_name[100];
       sprintf(file_name, "objects_info_%d.txt", buf_num + 2);
       FILE *fp = fopen(file_name, "rt");
       double max_dist = 1e10;
       int box_num;
       fscanf(fp, "%d", &box_num);

       for(int n = 0; n < box_num; n ++)
       {
              int x1, y1, x2, y2;
              fscanf(fp, "%d %d %d %d", &x1, &x2, &y1, &y2);
              for(int i = 0; i < 64; i++)
              {
                   float tmp;
                   fscanf(fp, "%f", &tmp);
              }
              pt1.x= x1;
              pt1.y= y1;
              pt2.x= x2;
              pt2.y= y2;
              cvRectangle(display_buf[cam_id][write_obuf_id][buf_num], pt1, pt2, CV_RGB(255, 255, 0), 1, CV_AA, 0 );
        }
        fclose(fp);
    }

}

void *read_obuf(void *idp)
{
    int i, m;
    for( i = 0; i < frame_buffer_len; i ++)
    {
        for( m = 0; m < cam_num; m++ )
        {
           char ss[100];
           sprintf(ss, "video%d", m+1);
           cvShowImage(ss, display_buf[m][read_obuf_id][i]);
        }
           cvWaitKey(60);
    }

    read_obuf_id = 1 - read_obuf_id;
    return 0;
}


int load_background_images()
{
    for( int i = 0; i < cam_num; i++)
       gray_background[i] = cvCreateImage(cvSize(width,height), IPL_DEPTH_8U, 1);

    IplImage * background_image;
    for(int k = 0; k < cam_num; k++)
    {
     char fn[100];
     sprintf(fn, "background%d.jpg", k+1);
     if( (background_image = cvLoadImage(fn, 1)) == 0 )
         return -1;
      cvCvtColor(background_image, gray_background[k], CV_BGR2GRAY);
      cvReleaseImage(&background_image);
    }
}

void release_background_images()
{
    for( int i = 0; i < cam_num; i++)
        cvReleaseImage(&gray_background[i]);
}


void track(int cam_id)
{
    int sequence_len = frame_buffer_len+1;
    int object_num;
    int m, n;

    FILE *fp = fopen("objects_info_1.txt", "rt");
    fscanf(fp, "%d", &object_num);
    fclose(fp);

    CvFont font;
    cvInitFont( &font, CV_FONT_HERSHEY_COMPLEX, 1, 1, 0.0, 2, CV_AA);
    char cmd[500];
    float res[3][frame_buffer_len][6];
    char fname[100];
    char ss[200];
    CvPoint pt1, pt2, ptx, pty;

    for(m = 0; m < object_num; m++)
    for(n = 0; n < sequence_len; n++)
    {
	     sprintf(cmd, "cp -f ./objects_info_%d.txt objects_info_h%d_%d.txt", n+1, m+1, n+1);
             system(cmd);
    }

    sprintf(cmd, "./detoverlap %d %d > ttt.txt", object_num, sequence_len);
    system(cmd);
    sprintf(cmd, "./lpgen %d %d > lp.lpt", object_num, sequence_len);
    system(cmd);
    system("./sollp --nomip --lpt lp.lpt -o out2.txt");
    sprintf(cmd, "./rounding %d %d > out.txt", object_num, frame_buffer_len);
    system(cmd);
    
    fp = fopen("out.txt", "rt");

    for(n = 0; n < object_num; n ++)
    {

              int  x1, y1, x2, y2;
              float z;
              int box_id;
              for(m = 0; m < frame_buffer_len; m++)
              {
                     fscanf(fp, "%d %d %d %d %f %d", &x1, &x2, &y1, &y2, &z, &box_id);
                     res[n][m][0] = x1;
                     res[n][m][1] = y1;
                     res[n][m][2] = x2;
                     res[n][m][3] = y2;
                     res[n][m][4] = z;
                     res[n][m][5] = box_id;
               }

     }

     fclose(fp);


     int frame_count = 0;
     for(int buf_num = 0; buf_num < frame_buffer_len; buf_num ++)
     {
      bool box_used[100];
      for(n = 0; n < 100; n++)
         box_used[n] = false;

      for(n = 0; n < object_num; n ++)
      {
			    
	int x1, y1, x2, y2;
	x1 = res[n][frame_count][0];
	y1 = res[n][frame_count][1];
	x2 = res[n][frame_count][2];
	y2 = res[n][frame_count][3];
        int box_id = res[n][frame_count][5];
        if (box_id > 99) box_id = 99;

        if( res[n][frame_count][4] < 0.5 && !box_used[box_id])
        {
	     pt1.x= x1;
             pt1.y= y1;
             pt2.x= x2;
             pt2.y= y2;
             box_used[box_id] = true;

             cvRectangle(display_buf[cam_id][write_obuf_id][buf_num], pt1, pt2, CV_RGB(255, 0, 0), 2, CV_AA, 0 );
	     sprintf(ss, "%d", n);
             cvPutText(display_buf[cam_id][write_obuf_id][buf_num], ss, pt1, &font, CV_RGB(0,0,255));
         }
				
       }
       frame_count ++;
     }

}


